#include <u.h>
#include <errno.h>
#include <stdio.h>
#include <libelf.h>
#include <libdwarf.h>
#include <dwarf.h>

#include "queue.h"
#include "dat.h"
#include "hash.h"
#include "type.h"

int		debug;
int		dosyms;
int		dotypes;
int		verbose;
int		endianness;
char		*argv0 = "dwarf2cqct";
Dwarf_Error	dwarferror;

static	int	processinput(char*, int);
static	int	elfwalk(int, const char*);
static	int	dwarfwalk(int, const char*, Elf*, int);
extern void fixbrokendwarfarraytypes (void);

static void
usage(void)
{
	fprintf(stderr, "usage: dwarf2cqct [-?v] [-st] [-u <uniq>] [fname]\n");
	exit(1);
}

int
main(int argc, char *argv[])
{
	int i;
	int fd, rv;
	char *fname;

	dwarfvarinit();
	dwarftypeinit();

	ARGBEGIN{
	case 'D':
		debug = 1;
		break;

	case 's':
		dosyms = 1;
		break;

	case 't':
		dotypes = 1;
		break;

	case 'u':
		anonuniq = EARGF(usage());
		break;

	case 'v':
		verbose++;
		break;

	case '?':
	default:
		usage();
		break;
	}ARGEND;

	if(!dosyms && !dotypes){
		dosyms = 1;
		dotypes = 1;
	}

	if(argc == 0){
		fd = 0;
		fname = "<stdin>";
		rv = processinput(fname, fd);
		if(rv < 0)
			goto Fail;
	}else{
		for(i=0; i<argc; i++){
			fname = argv[i];
			fd = open(fname, O_RDONLY);
			if(fd < 0){
				fprintf(stderr, "can't open '%s': %s\n",
					fname, strerror(errno));
				exit(1);
			}
			rv = processinput(fname, fd);
			close(fd);
			if(rv < 0)
				goto Fail;
		}
	}
	return 0;

  Fail:
	fprintf(stderr, "failure\n");
	return 1;
}

static int
processinput(char *fname, int fd)
{
	int rv;

	rv = elfwalk(fd, fname);
	if(rv < 0)
		goto Fail;
	rv = dwarfresolvetypes();
	if(rv < 0)
		goto Fail;
	fixbrokendwarfarraytypes ();
	if(dotypes)
		dwarfprinttypes();
	if(dosyms){
		rv = dwarfresolvesyms();
		if(rv < 0)
			goto Fail;
		dwarfprintsyms();
	}
	return 0;

  Fail:
	return -1;
}

static int
elfwalk(int fd, const char *fname)
{
	Elf *ar;
	Elf *elf;
	Elf_Cmd cmd;
	Elf_Scn *scn;
	Elf32_Shdr *sh32;
	Elf64_Shdr *sh64;
	Elf32_Ehdr *eh32;
	Elf64_Ehdr *eh64;

	int nscn, i;
	int nm, sz, rv;
	int archive, error;

	char *secname;
	uchar elfdata;

	error = 0;
	archive = 0;
	eh32 = nil;
	eh64 = nil;
	cmd = ELF_C_READ;

	elf_version(EV_NONE);
	if(elf_version(EV_CURRENT) == EV_NONE){
		fprintf(stderr, "error: libelf busted?\n");
		return -1;
	}

	ar = elf_begin(fd, cmd, nil);
	if(elf_kind(ar) == ELF_K_AR)
		archive = 1;
	while(!error){
		elf = elf_begin(fd, cmd, ar);
		if(elf == nil)
			break;
		eh32 = elf32_getehdr(elf);
		if(eh32 != nil)
			elfdata = eh32->e_ident[EI_DATA];
		else{
			eh64 = elf64_getehdr(elf);
			if(eh64 == nil)			/* XXX: quiet */
				goto Next;
			elfdata = eh64->e_ident[EI_DATA];
		}
		switch(elfdata){
		case ELFDATA2LSB:
			endianness = LittleEndian;
			break;

		case ELFDATA2MSB:
			endianness = BigEndian;
			break;

		default:
			endianness = NoEndian;
			break;
		}
		if(eh32 != nil)
			nscn = eh32->e_shnum;
		else
			nscn = eh64->e_shnum;
		for(i=0; verbose && i<nscn; i++){
			scn = elf_getscn(elf, i);
			if(eh32 != nil){
				sh32 = elf32_getshdr(scn);
				nm = sh32->sh_name;
				sz = sh32->sh_size;
				secname = elf_strptr(elf, eh32->e_shstrndx, nm);
			}else{
				sh64 = elf64_getshdr(scn);
				nm = sh64->sh_name;
				sz = sh64->sh_size;
				secname = elf_strptr(elf, eh64->e_shstrndx, nm);
			}
			fprintf(stderr, "section %s size %d\n", secname, sz);
		}
		rv = dwarfwalk(fd, fname, elf, archive);
		if(rv < 0)
			error = 1;	
	Next:
		cmd = elf_next(elf);
		elf_end(elf);
	}
	elf_end(ar);
	if(endianness == NoEndian)
		error = 1;
	if(error)
		return -1;
	return 0; 
}

static int
dwarfwalk(int fd, const char *fname, Elf *elf, int archive)
{
	int rv;
	Dwarf_Debug dbg;

	rv = dwarf_elf_init(elf, DW_DLC_READ, nil, nil, &dbg, &dwarferror);
	if(rv == DW_DLV_NO_ENTRY)
		return 0;
	if(rv != DW_DLV_OK){
		fprintf(stderr, "error: %s\n", dwarf_errmsg(dwarferror));
		return -1;
	}
	/* XXX: set initial frame rule value and frame rule table size */
	dwarfinfowalk(dbg);
	return 0;
}
