#include <u.h>
#include <stdio.h>
#include <libelf.h>
#include <libdwarf.h>
#include <dwarf.h>

#include "dat.h"

void
dwarfattrvalueprint(Dwarf_Debug dbg, DwarfAttrValue *val)
{
	switch(val->form){
	case DW_FORM_addr:
		fprintf(stderr, "0x%llx", val->addr);
		break;

	case DW_FORM_ref_addr:
	case DW_FORM_ref1:
	case DW_FORM_ref2:
	case DW_FORM_ref4:
	case DW_FORM_ref8:
	case DW_FORM_ref_udata:
		fprintf(stderr, "%llu", val->off);
		break;

	case DW_FORM_data1:
	case DW_FORM_data2:
	case DW_FORM_data4:
	case DW_FORM_data8:
	case DW_FORM_sdata:
	case DW_FORM_udata:
		if(val->issigned)
			fprintf(stderr, "%lld", val->sdata);
		else
			fprintf(stderr, "%llu", val->udata);
		break;

	case DW_FORM_strp:
	case DW_FORM_string:
		fprintf(stderr, "%s", val->str);
		break;

	default:
		if(verbose)
			fprintf(stderr, "unhandled form '%s'\n",
				dwarfformname(val->form));
		break;
	}

	return;
}

/*
 * String objects do not need to be freed.
 */
void
dwarfattrvalueclear(Dwarf_Debug dbg, DwarfAttrValue *val)
{
	if(val == nil)
		return;

	switch(val->form){
	case DW_FORM_block:
	case DW_FORM_block1:
	case DW_FORM_block2:
	case DW_FORM_block4:
		if(val->block);
			dwarf_dealloc(dbg, val->block, DW_DLA_BLOCK);
		break;

	default:
		break;
	}

	return;
}

int
dwarfattrvalue(Dwarf_Debug dbg, Dwarf_Die die, Dwarf_Attribute attr, Dwarf_Half atag, DwarfAttrValue *val, DwarfCompileUnit *cu)
{
	int rv;
	Dwarf_Bool flag;
	Dwarf_Half dattr;

	rv = dwarf_whatform(attr, &val->form, &dwarferror);
	if(rv == DW_DLV_ERROR){
		fprintf(stderr, "error: unknown form for attribute %d\n", atag);
		goto Error;
	}else if(rv == DW_DLV_NO_ENTRY)
		goto Done;
	if(verbose > 1)
		fprintf(stderr, "form %s\n", dwarfformname(val->form));
	switch(val->form){
	case DW_FORM_flag:
		rv = dwarf_formflag(attr, &flag, &dwarferror);
		if(rv != DW_DLV_OK)
			goto Error;
		val->udata = flag;
		break;

	case DW_FORM_addr:
		rv = dwarf_formaddr(attr, &val->addr, &dwarferror);
		if(rv != DW_DLV_OK){
			fprintf(stderr, "error: addr form with no addr?\n");
			goto Error;
		}
		break;

	case DW_FORM_ref_addr:
		rv = dwarf_global_formref(attr, &val->off, &dwarferror);
		if(rv != DW_DLV_OK){
			fprintf(stderr, "error: ref addr form with no ref?\n");
			goto Error;
		}
		break;

	case DW_FORM_ref1:
	case DW_FORM_ref2:
	case DW_FORM_ref4:
	case DW_FORM_ref8:
	case DW_FORM_ref_udata:
		rv = dwarf_formref(attr, &val->off, &dwarferror);
		if(rv != DW_DLV_OK){
			fprintf(stderr, "error: ref form with no ref?\n");
			goto Error;
		}
		break;

	case DW_FORM_block:
	case DW_FORM_block1:
	case DW_FORM_block2:
	case DW_FORM_block4:
		rv = dwarf_formblock(attr, &val->block, &dwarferror);
		if(rv != DW_DLV_OK){
			fprintf(stderr, "error: block form with no data?\n");
			goto Error;
		}
		break;

	case DW_FORM_data1:
	case DW_FORM_data2:
	case DW_FORM_data4:
	case DW_FORM_data8:
		rv = dwarf_whatattr(attr, &dattr, &dwarferror);
		if(rv == DW_DLV_ERROR || rv == DW_DLV_NO_ENTRY){
			fprintf(stderr, "error: buggered data form\n");
			goto Error;
		}
		val->issigned = 0;
		rv = dwarf_formudata(attr, &val->udata, &dwarferror);
		if(rv != DW_DLV_OK){
			val->issigned = 1;
			rv = dwarf_formsdata(attr, &val->sdata, &dwarferror);
			if(rv != DW_DLV_OK){
				fprintf(stderr, "error: buggered data form\n");
				goto Error;
			}
		}
		break;

	case DW_FORM_sdata:
		val->issigned = 1;
		rv = dwarf_formsdata(attr, &val->sdata, &dwarferror);
		if(rv != DW_DLV_OK){
			fprintf(stderr, "error: buggered sdata form\n");
			goto Error;
		}
		break;

	case DW_FORM_udata:
		val->issigned = 0;
		rv = dwarf_formudata(attr, &val->udata, &dwarferror);
		if(rv != DW_DLV_OK){
			fprintf(stderr, "error: buggered udata form\n");
			goto Error;
		}
		break;

	case DW_FORM_strp:
	case DW_FORM_string:
		val->str = nil;
		rv = dwarf_formstring(attr, &val->str, &dwarferror);
		if(rv == DW_DLV_OK || DW_DLV_NO_ENTRY) {}
		else{
			fprintf(stderr, "error: borked string form\n");
			goto Error;
		}
		break;

	default:
		if(verbose)
			fprintf(stderr, "unhandled form %d/%s\n", val->form,
				dwarfformname(val->form));
		goto Error;
		break;
	}

	val->isvalid = 1;

  Done:
	return 0;

  Error:
	return -1;
}
