#include <u.h>
#include <stdio.h>
#include <dwarf.h>
#include <libdwarf.h>

#include "queue.h"
#include "dat.h"
#include "hash.h"
#include "type.h"
#include "util.h"

extern char * get_cbname (unsigned int);

static	int	needswrap(Type*);
static	int	needsparen(unsigned);

static	char	*fmtdecllist(Type*);
static	char	*fmttype(Type*, char*);
static	char	*fmttypedecl(Type*, char*);
static	char	*fmtqual(Type*, const char*, char*);
static	char	*fmtsumember(const Decl*, char*, int);

static int
needswrap(Type *t)
{
	switch(t->type){
	case Tarr:
	case Tfun:
		return 1;

	case Tconst:
	case Tvolatile:
	case Trestrict:
	case Ttypedef:
	case Tptr:
	case Tref:
		return needswrap(t->sub);
	}
	return 0;
}

static int
needsparen(unsigned type)
{
	switch(type){
	case Tfun:
	case Tarr:
	case Tref:
		return 1;

	default:
		return 0;
	}
}

#undef OPT
#define OPT(o)	(o)?" ":"", (o)?(o):""

static char*
fmtdecllist(Type *t)
{
	int m;
	Decl *d;
	char *buf;
	char *t0, *t1;

	buf = strdup("");
	STAILQ_FOREACH(d, &t->members, link){
		t0 = nil;
		if(d->name && needswrap(d->type))
			t0 = strdup(d->name);
		t0 = fmttype(d->type, t0);
		if(d->name && !needswrap(d->type)){
			m = strlen(t0)+1+strlen(d->name)+1;
			t1 = malloc(m);
			snprintf(t1, m, "%s %s", t0, d->name);
			free(t0);
			t0 = t1;
		}
		buf = estrcat(buf, t0);
		free(t0);
		if(STAILQ_NEXT(d, link))
			buf = estrcat(buf, ", ");
	}
	return buf;
}

/* XXX: printing of qualifiers is buggered, so we just don't bother with it */
static char*
fmtqual(Type *t, const char *q, char *o)
{
	char *s;

	s = fmttype(t, o);
	return s;
#ifdef NOTDEF
	int m;
	char *s;
	char *buf;

	s = fmttype(t, o);

	m = strlen(q)+1+strlen(s)+1+2;
	if(o != nil)
		m += strlen(o)+1;
	buf = malloc(m);
	bzero(buf, m);
	snprintf(buf, m, "%s", s);
	free(s);
	return buf;
#endif
}

static char*
fmttype(Type *t, char *o)
{
	int i, ty;
	unsigned m, no;
	char *buf, *w, *pl;
	char tmpbuf[64];
	static char *refcomment = " /* was & */";

	no = 0;
	if(o)
		no = strlen(o);

	ty = t->type;

	switch(ty){
	case Tplastic:
		m = strlen("<plastic>")+1+no+1;
		buf = malloc(m);
		snprintf(buf, m, "<plastic>%s%s", OPT(o));
		free(o);
		return buf;

	case Tbase:
		m = strlen(t->tag)+1+no+1;
		buf = malloc(m);
		snprintf(buf, m, "%s%s%s", t->tag, OPT(o));
		free(o);
		return buf;

	case Ttypedef:
		m = strlen(t->tag)+1+no+1;
		buf = malloc(m);
		snprintf(buf, m, "%s%s%s", t->tag, OPT(o));
		free(o);
		return buf;

	case Tstruct:
	case Tunion:
		w = t->type == Tstruct ? "struct" : "union";
		if(t->tag){
			m = strlen(w)+1+strlen(t->tag)+1+no+1;
			buf = malloc(m);
			snprintf(buf, m, "%s %s%s%s", w, t->tag, OPT(o));
		}else{
			m = strlen(w)+1+no+1;
			buf = malloc(m);
			snprintf(buf, m, "%s%s%s", w, OPT(o));
		}
		free(o);
		return buf;

	case Tenum:
		if(t->tag){
			m = 4+1+strlen(t->tag)+1+no+1;
			buf = malloc(m);
			snprintf(buf, m, "enum %s%s%s", t->tag, OPT(o));
		}else{
			m = 4+1+no+1;
			buf = malloc(m);
			snprintf(buf, m, "enum%s%s", OPT(o));
		}
		free(o);
		return buf;

	case Tptr:
		m = 2+no+1+1;
		buf = malloc(m);
		if(needsparen(t->sub->type))
			snprintf(buf, m, "(*%s)", o?o:"");
		else
			snprintf(buf, m, "*%s", o?o:"");
		free(o);
		return fmttype(t->sub, buf);

	case Tref:
		m = 2+no+1+1+strlen(refcomment);
		buf = malloc(m);
		if(needsparen(t->sub->type))
			snprintf(buf, m, "(*%s)%s", o?o:"", refcomment);
		else
			snprintf(buf, m, "*%s%s", o?o:"", refcomment);
		free(o);
		return fmttype(t->sub, buf);

	case Tarr:
		buf = o;
		if(buf == nil)
			buf = strdup("");
		for(i=0; i<t->narr; i++){
			m = t->dim[i];
			snprintf(tmpbuf, sizeof tmpbuf, "[0x%x]", m);
			buf = estrcat(buf, tmpbuf);
		}
		if(t->narr == 0)
			buf = estrcat(buf, "[]");
		return fmttype(t->sub, buf);

	case Tfun:
		assert(o != nil);
		pl = fmtdecllist(t);
		m = no+1+strlen(pl)+1+1;
		buf = malloc(m);
		snprintf(buf, m, "%s(%s)", o, pl);
		free(o);
		free(pl);
		return fmttype(t->sub, buf);

	case Tvararg:
		return strdup("...");

	case Tconst:
		return fmtqual(t->sub, "const", o);

	case Tvolatile:
		return fmtqual(t->sub, "volatile", o);

	case Trestrict: 
		return fmtqual(t->sub, "restrict", o);

#ifdef NOTDEF
	default:
		fatal("bad declaration type");
#endif
	}
	return nil;
}

static char*
fmttypedecl(Type *t, char *o)
{
	int m;
	Decl *d;
	int xtab, no;
	char *t0, *t1;
	char *buf, *s;
	char tmpbuf[64];

	s = nil;
	buf = o;
	no = 0;
	if(o)
		no = strlen(o);

	switch(t->type){
	case Tplastic:
		fprintf(stderr, "encountered plastic type\n");
		break;

	case Tfun:
	case Tarr:
	case Tptr:
	case Tref:
	case Tbase:
	case Tconst:
	case Trestrict:
	case Tvolatile:
		/*
		 * Omit these types at the top level.
		 */
		if(debug && verbose)
			fprintf(stderr, "omit %s\n", typetab[t->type]);
		break;

	case Tenum:
		buf = fmttype(t, nil);
		buf = estrcat(buf, " {\n");
		STAILQ_FOREACH(d, &t->members, link){
			if(d->value.str){
				m = 1+strlen(d->name)+3+strlen(d->value.str)+1;
				t0 = malloc(m);
				snprintf(t0, m, "\t%s = %s,\n",
					d->name, d->value.str);
			}else{
				m = 1+strlen(d->name)+3+sizeof(tmpbuf)+1;
				t0 = malloc(m);
				if(d->value.issigned)
					snprintf(t0, m, "\t%s = %ld,\n",
						d->name,
						(long)d->value.sdata);
				else
					snprintf(t0, m, "\t%s = %lu,\n",
						d->name,
						(ulong)d->value.udata);

			}
			buf = estrcat(buf, t0);
			free(t0);
		}
		snprintf(tmpbuf, sizeof tmpbuf, "};\n");
		buf = estrcat(buf, tmpbuf);
		break;

	case Tstruct:
	case Tunion:
		buf = fmttype(t, nil);
		if(t->flags & Fdecl){
			buf = estrcat(buf, ";");
			break;
		}
		buf = estrcat(buf, " {\n");

		xtab = 0;
		STAILQ_FOREACH(d, &t->members, link)
			if(d->bitsize)
				xtab = 1;
		STAILQ_FOREACH(d, &t->members, link){
			if(d->type == nil){
				fprintf(stderr, "missing type for member '%s'\n", d->name);
				continue;
			}
			buf = fmtsumember(d, buf, xtab);
		}
		snprintf(tmpbuf, sizeof tmpbuf, "\t@0x%x;\n};\n", t->size);
		buf = estrcat(buf, tmpbuf);
		break;

	case Ttypedef:
	  if (t && !t->sub && t->encoding) {
	    /* Then we have an internally generated typedef for
	       a nonstandard base type without a type reference.
	       The encoding gives us the correct reference even
	       though it does not appear in the DWARF info for
	       the compilation unit containing this tag */
	    t0 = get_cbname (t->encoding);
	    m = 7+1+strlen(t0)+1+strlen(t->tag)+1+1+no+1;
	    buf = malloc(m);
	    snprintf(buf, m, "typedef %s %s;%s%s", t0, t->tag, OPT(o));
	  }
	  else if(needswrap(t)){
	    t1 = strdup(t->tag);
	    t0 = fmttype(t->sub, t1);
	    m = 7+1+strlen(t0)+1+no+1;
	    buf = malloc(m);
	    snprintf(buf, m, "typedef %s;%s%s", t0, OPT(o));
	  }else{
	    t0 = fmttype(t->sub, nil);
	    m = 7+1+strlen(t0)+1+strlen(t->tag)+1+1+no+1;
	    buf = malloc(m);
	    snprintf(buf, m, "typedef %s %s;%s%s", t0, t->tag, OPT(o));
	    free(t0);
	  }
	  break;
	  
	default:
	  fprintf(stderr, "fmtdecl unhandled type %s %d\n", typetab[t->type], t->flags);
	  break;
	}
	return buf;
}

static char*
fmtsumember(const Decl *d, char *buf, int xtab)
{
	char *t0, *t1;
	char tmpbuf[64];

	if(d->bitsize == 0)
		snprintf(tmpbuf, sizeof tmpbuf, "\t@0x%lx%s\t",
			(ulong)d->offset, xtab?"\t":"");
	else
		snprintf(tmpbuf, sizeof tmpbuf, "\t@@(8*0x%lx+%u)\t",
			(ulong)d->offset, d->bitoffset);
	buf = estrcat(buf, tmpbuf);
	t0 = strdup(d->name);
	t1 = fmttype(d->type, t0);	/* frees t0 */
	buf = estrcat(buf, t1);
	if(d->bitsize){
		snprintf(tmpbuf, sizeof tmpbuf, " : %u", (unsigned)d->bitsize);
		buf = estrcat(buf, tmpbuf);
	}
	snprintf(tmpbuf, sizeof tmpbuf, ";\n");
	buf = estrcat(buf, tmpbuf);
	free(t1);

	return buf;
}

void
printtype(Type *t)
{
	char *s;
	DwarfType *dt;

	if(t->flags & Fmark)
		return;
	
	t->flags |= Fmark;

	s = fmttypedecl(t, nil);
	if(s && *s){
		if(verbose > 1){
			dt = type2dwarftype(t);
			if(dt && dt->file)
				printf("%c* %s:%lu *%c\n",
					'/', dt->file, dt->line, '/');
		}
		printf("%s\n", s);
		free(s);
	}
}

void
printsym(DwarfVar *v)
{
	char *s;
	Type *t;

	if(v->decl->type){
		/* Synthesize pointer types */
		t = malloc(sizeof *t);
		t->type = Tptr;
		t->sub = v->decl->type;
		switch(v->decl->type->type){
		case Tptr:
		case Tarr:
			s = fmttype(v->decl->type, v->decl->name);
			break;

		default:
//			s = fmttype(t, v->decl->name);
			s = fmttype(v->decl->type, v->decl->name);
			break;
		}
		if(s && *s)
			printf("@0x%llx\t%s;\n", v->addr, s);
		free(s);
		free(t);
	}else
		printf("@0x%llx\tvoid\t*%s;\t/* No type */\n",
			v->addr, v->decl->name);
	return;
}
